#!/usr/bin/env ruby

# 1230, Fri 4 Jun 10 (NZST)
#
# mk-owt-config.rb: Makes 'default' Ruby config file
#   User will probably need to edit the owt-config file  <<<
#   so that it works in their local owtmon environment   <<<
#
# Copyright 2010, Nevil Brownlee, U Auckland | CAIDA

cfname = 'owt-config.rb'
if FileTest.exists?(cfname) && FileTest.size(cfname) != 0
  print "Existing config file #{cfname} already exists, leaving it unchanged"
  exit
end

def p_exists?(fn)
  $LOAD_PATH.each {|path| return true if File.exist?(path + '/' + fn) }
  return true if File.exist?('./' + fn) 
  false
end
if p_exists?('IPprefix.bundle') || p_exists?('IPprefix.so')
  require 'IPprefix'
else
  print "Need IPprefix so we can generate a config file!"
  exit
end

cf = File.new(cfname, 'w')
conf_hdr = <<END
# This file is a default version, generated by ./configure
#   >>> You probably need to edit it to make specify your   <<<
#   >>> local owtmon environment correctly                  <<< 

# Ruby owtmon programs use this config file by saying
#   require "#{File.dirname($0)}/owt-config"


# Default command-line parameters for owtmon . . .

SourceURI = 'pcapint:eth0'  # (libtrace-style) URI for input data source

SummaryDir = 'sum2'    # Directory where summary files are kept
                       #   (in YYYY-MM-DD subdirs) 

StatsFile = 'owt-stats'  # Statistics file for current day's summaries


# System constants . . .

SumFileType = 'sum2'   # Filetype for summary files

Hours_reqd = 1         # Summary file interval (hours)

MxLifetime = 172800    # Upper bound for lifetime distribs (seconds in 2 days)

N_Streams =   300000   # Memory allocation: number of one-way sources;
                       #    owtmon allocates blocks this size as needed

N_PktDatas = 4000000   # Memory allocation: number of packets for 
                       #   not-yet-analysed sources

EnoughPkts = 20        # Normal number of packets needed to classify a source

MinPkts = 10           # Minimum number of packets to classify a source
                       #   (at end of a summary interval)

N_ProbePorts = 100     # Number of TCP and UDP probe ports kept in summaries

MeterLoc = 'Unknown'   # Normally use 3-letter airport code, e.g. SAN, AKL


# Network information . . .

END
cf.print conf_hdr

Home = Dir.pwd
cf.print "Home = '#{Home}'  # owtmon home directory\n\n"

hostname = `hostname`.chomp
Host = hostname.split('.')[0]
cf.print "Host = '#{Host}'\n\n"

require 'socket'
addr = Socket.gethostbyname(hostname)
af = addr[2];  abytes = addr[3]
LocalAddress = IPprefix.new(af == Socket::AF_INET6 ? 6 : 4, abytes,
  af == Socket::AF_INET6 ? 48 : 16)

cf.print "LocalAddress = '#{LocalAddress}'\n"
cf.print "      # This is the IP address prefix(es) for the network that\n"
cf.print "      #   your one-way packets' destination address fall within.\n"
cf.print "      # It may be a comma-separated list of mixed IPv4 and/or\n"
cf.print "      #   IPv6 addresses.  White space is ignored after a comma.\n"
cf.print "\n"

cf.close

print "New config file #{cfname} generated, edit it for your local conditions"
