/* 1313, Wed 19 May 10 (NZST)
   1955, Thu 22 Apr 10 (PDT)
   1829, Sun  9 Aug 09 (NZST)

   owt_monitor.c: Ruby C interface to owtmon

   Copyright (C) 2007-2010 by Nevil Brownlee, U Auckland | CAIDA | WAND */

#include "owt.h"
#include "ruby.h"

static VALUE IPprefix; /* IPprefix class */

static VALUE mSE;      /* OWTmonitor module */
static VALUE cST;      /* Monitor class */
static VALUE cStream;  /* OWTevent class */
static VALUE cSEmsg;   /* OWTmessage class */

static struct opq_entry cs_oqe;

static VALUE cs_init(VALUE self, VALUE s_id) {
   rb_iv_set(self, "@id", s_id);
   return self;
   }

static VALUE cs_last_tsec(VALUE self) {
   struct stream *tsp;
   uint32_t ts;
   cs_oqe.s_id = FIX2INT(rb_iv_get(self, "@id"));
   tsp = stream_from_oqe(&cs_oqe);
   mutex_lock("cts", &stream_data_mutex);
   ts = (uint32_t)(tsp->last_us/1000000.0);
   mutex_unlock("cts", &stream_data_mutex);
   return INT2NUM(ts);
   }

static VALUE cs_ip_addresses(VALUE self) {
   struct stream *tsp;
   uint8_t *a;
   VALUE pa[3];
   VALUE addrs[2];  /* Result: [src_addr, dest_addr] */

   cs_oqe.s_id = FIX2INT(rb_iv_get(self, "@id"));
   tsp = stream_from_oqe(&cs_oqe);
   if (tsp->version == 4) {
      pa[0] = INT2FIX(4);  pa[2] = INT2FIX(32);
      a = tsp->src_addr.a.a;
      pa[1] = rb_str_new((char *)a, IP4_ADDR_LEN);
      addrs[0] = rb_class_new_instance(3, pa, IPprefix);
      }
   else {  /* v6 */
      pa[0] = INT2FIX(6);  pa[2] = INT2FIX(128);
      a = tsp->src_addr.a.a;
      pa[1] = rb_str_new((char *)a, IP6_ADDR_LEN);
      addrs[0] = rb_class_new_instance(3, pa, IPprefix);
      }
   pa[1] = Qnil;
   return rb_ary_new4(2, addrs);
   }

static VALUE cs_usage(VALUE self) {
   struct opq_entry cs_oqe;
   struct stream *sp;
   VALUE ra[4];  /* first_fwd, src_addr, tot_pkts, lifetime_us */
   VALUE pa[2];
   cs_oqe.s_id = FIX2INT(rb_iv_get(self, "@id"));
   sp = stream_from_oqe(&cs_oqe);

   pa[0] = INT2FIX(sp->version);
   if (sp->version == 4) {
      pa[1] = rb_str_new((char *)sp->src_addr.a.a, 4);
      }
   else pa[1] = rb_str_new((char *)sp->src_addr.a.a, 16);
   ra[0] = sp->first_fwd ? Qtrue : Qfalse;
   ra[1] = rb_class_new_instance(2, pa, IPprefix);
   ra[2] = INT2FIX(sp->tp + sp->fp);
   ra[3] = rb_float_new(sp->last_us - sp->first_us);
   return rb_ary_new4(4, ra);
   }

static int get_pkt_attribs(struct pkt_data *pktp)
{  /* Return 0 to continue, 1 to stop walking */
   VALUE ra[9];
      /* arrival_time, proto, src_port, dst_port, dst_addr 
            ip_len, tcp_flags, cfc */
   VALUE pa[2];
   char *pkk, pkt_key[90];  int n;
   if (pktp->dst_addr.ver == 4) {
      n = snprintf(pkt_key, sizeof(pkt_key), "%u.%u.%u.%u",
         pktp->dst_addr.a.a[0],pktp->dst_addr.a.a[1],
	 pktp->dst_addr.a.a[2],pktp->dst_addr.a.a[3]);
      pkk = &pkt_key[n];
      }
   else pkk = v6addr_to_s(pkt_key, &pktp->dst_addr);

   snprintf(pkk, sizeof(pkt_key)-48, " %u %u %u %u %u %u",
      pktp->src_port, pktp->dst_port, pktp->proto,
      pktp->ip_len, pktp->tcp_flags, pktp->cfc);
   ra[0] = rb_str_new(pkt_key, strlen(pkt_key));

   ra[1] = rb_float_new(pktp->arrival_time);
   ra[2] = INT2FIX(pktp->proto);
   ra[3] = UINT2NUM(pktp->src_port);
   ra[4] = UINT2NUM(pktp->dst_port);
      pa[0] = INT2FIX(pktp->dst_addr.ver);
      pa[1] = rb_str_new((char *)pktp->dst_addr.a.a,
         pktp->dst_addr.ver == 4 ? 4 : 16);
   ra[5] = rb_class_new_instance(2, pa, IPprefix);
   ra[6] = INT2FIX(pktp->ip_len);
   ra[7] = INT2FIX(pktp->tcp_flags);
   ra[8] = INT2FIX(pktp->cfc);

   rb_yield(rb_ary_new4(sizeof(ra)/sizeof(VALUE), ra));
   return 0;
   }

static VALUE cs_each_pkt(VALUE self) {  /* pkt_datas for src_host self */
   struct opq_entry cs_oqe;
   struct stream *sp;
   struct pkt_data *pktp;

   cs_oqe.s_id = FIX2INT(rb_iv_get(self, "@id"));
   sp = stream_from_oqe(&cs_oqe);

   pkt_data_q_lock(&sp->pktq);
   if (sp->pktq.length != 0)
      pkt_data_walk_down(&sp->pktq, get_pkt_attribs);
   pkt_data_q_unlock(&sp->pktq);
   return self;
   }

static clear_pkt_queue(struct stream *sp) {
   /* Also counts the dest addresses for this source's type */
   struct pkt_data *pktp;
   int st = sp->src_type;

   pkt_data_q_lock(&sp->pktq);
   while (pktp = pkt_data_shift(&sp->pktq)) {
#if MKDESTCOUNTS  /* Only care about IPv4 for now */
      if (pktp->dst_addr.ver == 4) {
         dest_addrs_per_source_type[st][
            dest_addr_bin(pktp->dst_addr.a.v4)] += 1;
         }
#endif
      free_pkt_data(pktp);  /* locks/unlocks pkt_data_q */
      }
   pkt_data_q_unlock(&sp->pktq);
   }

static VALUE cs_set_source_attribs(VALUE self, VALUE v_aa) {
   VALUE v_arg;
   struct stream *sp;

   cs_oqe.s_id = FIX2INT(rb_iv_get(self, "@id"));
   sp = stream_from_oqe(&cs_oqe);
#if CHECKPKTQUEUE
   pkt_data_queue_ok(&sp->pktq, 1, 201);  //dd
#endif

   if (sp->state == SS_TWO_WAY)  /* ENOUGH -> TWO_WAY */
      return self;  /* Let it time out */

   if ((v_arg = rb_ary_shift(v_aa)) != Qnil) {  /* src_type */
      sp->src_type = NUM2INT(v_arg);
      }
   v_arg = rb_ary_shift(v_aa);  /* probe_port (may be nil) */
   if (v_arg == Qnil) sp->probe_port_valid = 0;
   else {
      sp->probe_port = NUM2INT(v_arg);  sp->probe_port_valid = 1;
      }
   if ((v_arg = rb_ary_shift(v_aa)) != Qnil)  /* mean_retry_time */
     sp->retry_time = NUM2DBL(v_arg);
   if ((v_arg = rb_ary_shift(v_aa)) != Qnil)  /* n_dest_ports */
      sp->n_dest_ports = NUM2INT(v_arg);
   if ((v_arg = rb_ary_shift(v_aa)) != Qnil)  /* pkt_len_mode */
      sp->pkt_len_mode = NUM2INT(v_arg);

#if CHECKPKTQUEUE
   pkt_data_queue_ok(&sp->pktq, 1, 210);  //dd
#endif

   sp->pkts_last_classify = sp->totp;
   if (sp->tp >= pkts_to_record)  /* Keep them if less, may try again */
      clear_pkt_queue(sp);
#if CHECKPKTQUEUE
   pkt_data_queue_ok(&sp->pktq, 1, 220);  //dd
#endif
   sp->analysed = 1;  sp->msg_queued = OQ_NULL;

#if MKSTATSCOUNTS  /* Now we know stream's src_type */
   sp->last_active_interval = stats_interval_nbr;
   active_sources_this_interval[sp->src_type] += 1;
   unique_hosts_this_interval[sp->src_type] += 1;
#endif
   return self;
   }

static VALUE msg_init(VALUE self, VALUE type, VALUE stream, VALUE text) {
   rb_iv_set(self, "@type", type);
   rb_iv_set(self, "@source", stream);
   rb_iv_set(self, "@text", text);
   return self;
   }

static char *interface_names[MXINTERFACES];  /* For start_meter() call */
static int if_names_sz, mx_events, mx_opq_entries,
   mx_ip_addresses, mx_pkt_datas, mx_streams,
   pw_q_sz, mx_stream_txts, mx_flow_watchers,
   mx_source_types;
   
static VALUE owtm_initialize(VALUE self, VALUE v_params) {
   /* Just call class init() method, i.e. owtm_init() */
   rb_funcall(self, rb_intern("init"), 1, v_params);
   return self;
   }

static VALUE owtm_init(int argc, VALUE *argv, VALUE self) {
   VALUE v_interfaces, v_live,  v_ip_addresses, v_pkt_datas, v_streams;
   VALUE v_ra, v_arg, rs_if;
   int rs, if_len, s4, x6, j;
   char *if_name;

   uid_from_cwd();  /* Set my_uid and se_uid (from ~/ directory) */

   mx_events = DFMXEVENTS;  /* Object memory sizes */
   mx_opq_entries = DFMXOPQENTRIES;
   pw_q_sz = DFMXPWQUEUESZ;
   mx_stream_txts = DFMXSTREAMTXTS;
   mx_flow_watchers = DFMXFLOWWATCHERS;
   n_local_addrs = 0;
   pkts_to_record = DFPKTSTORECORD;  /* Declared in ut.h */
   min_pkts_to_classify = 10;
   min_pkts_for_active_source = 5;  /* Don't keep if <= 5 */

   traces_to_process = DFTRACESTOPROCESS;

   rs = rb_scan_args(argc, argv, "4*",  /* Get args (from Streamer.new) */
      &v_interfaces, &v_live, &v_pkt_datas, &v_streams,
      &v_ra);
   if_names_sz = RARRAY(v_interfaces)->len;
   if (if_names_sz > MXINTERFACES)
      log_msg(LOG_ERR, 27, "More than %d interfaces specified !!!");
   for (j = 0; j != if_names_sz; ++j) {
      rs_if = rb_ary_shift(v_interfaces);
      if_name = RSTRING(rs_if)->ptr;  if_len = RSTRING(rs_if)->len;
      interface_names[j] = malloc(if_len+1);
      strncpy(interface_names[j], if_name, if_len+1);
      }
   live_source = v_live == Qtrue;
   mx_pkt_datas = NUM2INT(v_pkt_datas);
   mx_streams = NUM2INT(v_streams);
   // mx_opq_entries = mx_streams;  /* 6 Aug 10 */

   if ((v_arg = rb_ary_shift(v_ra)) != Qnil)
     pkts_to_record = NUM2INT(v_arg);
   if ((v_arg = rb_ary_shift(v_ra)) != Qnil)
     time_rec_interval = NUM2INT(v_arg);
   if ((v_arg = rb_ary_shift(v_ra)) != Qnil)
     min_pkts_to_classify = NUM2INT(v_arg);
   if ((v_arg = rb_ary_shift(v_ra)) != Qnil)
      mx_source_types = NUM2INT(v_arg);
   if ((v_arg = rb_ary_shift(v_ra)) != Qnil)
      traces_to_process = NUM2INT(v_arg);

   logname = "owt.log";
   return self;
   }

static VALUE owtm_info(VALUE self) {
   VALUE ra[2];
   ra[0] = rb_str_new2(METER_VERSION);
#if STREAMTIMEOUTS
   ra[1] = Qtrue;
#else
   ra[1] = Qfalse;
#endif
   return rb_ary_new4(sizeof(ra)/sizeof(VALUE), ra);
   }

static VALUE owtm_log_msg(VALUE self, VALUE rdie, VALUE rmsg) {
   log_msg(LOG_WARNING, NUM2INT(rdie), RSTRING(rmsg)->ptr);
   return self;
   }

static uint32_t ipv4_dest_addr_b2(uint32_t a) {
   return (ntohl(a) >> 16) & 0x000000FF;  /* 2nd byte */
   }

static uint32_t ipv4_dest_addr_b3(uint32_t a) {
   return (ntohl(a) >> 8) & 0x000000FF;  /* 3nd byte */
   }

static VALUE owtm_set_local_addrs(VALUE self, VALUE v_addr) {
   VALUE v_prefix;
   int v, nb, len;
   uint8_t *ap;
   for (n_local_addrs = 0;;) {
      v_prefix = rb_ary_shift(v_addr);
      if (v_prefix == Qnil) 
	 return INT2FIX(n_local_addrs);
      if (n_local_addrs == MX_LOCAL_ADDRS) rb_raise(rb_eArgError,
         "set_local_addrs(): MX_LOCAL_ADDRS is %d", MX_LOCAL_ADDRS);
      v =  FIX2INT(rb_iv_get(v_prefix, "@version"));
      nb = v == 4 ? IP4_ADDR_LEN : IP6_ADDR_LEN;
      ap = (uint8_t *)RSTRING_PTR(rb_iv_get(v_prefix, "@addr"));
      len = FIX2INT(rb_iv_get(v_prefix, "@length"));
      memcpy(la[n_local_addrs].a.a, ap, nb);
      la[n_local_addrs].len = len;
      la[n_local_addrs].ver = v;
      if (v == 4) {
         if (len == 8) dest_addr_bin = ipv4_dest_addr_b2;
         else if (len == 16) dest_addr_bin = ipv4_dest_addr_b3;
         }
      n_local_addrs += 1;
      }
   return self;
   }

static VALUE owtm_set_uris(VALUE self, VALUE u_a) {
   other_uris = (char **)malloc((traces_to_process-1)*sizeof(char *));
   VALUE u_s;  int j;
   for (j = 0;  j != traces_to_process-1; ++j) {
      u_s = rb_ary_shift(u_a);
      other_uris[j] = malloc(RSTRING(u_s)->len+2);
      strcpy(other_uris[j], RSTRING(u_s)->ptr);
      }
   next_uri_nbr = 0;
   return self;
   }

static VALUE owtm_open(VALUE self) {
   start_meter(interface_names, if_names_sz, mx_events, mx_opq_entries, 
      mx_pkt_datas, mx_streams,
      pw_q_sz, mx_stream_txts, mx_flow_watchers,
      mx_source_types, traces_to_process);
   return UINT2NUM(start_ts);
   }

static VALUE owtm_close(VALUE self) {
   request_shutdown = 1;
   return self;
   }

static VALUE owtm_nodename(VALUE self) {
   return rb_str_new2(node.nodename);
   }

static VALUE owtm_each(VALUE self) {
   struct opq_entry *oqe;
   char buf[200];
   VALUE attribs[3];  /* SEmessage: @type, @id, @text */
   VALUE sa[1];       /* Stream: @id */

   for (;;) {
      t_semaphore_down(&opq_ready);  /* Wait for a full block */

      opq_entry_q_lock(&opq);
      oqe = opq_entry_shift(&opq);
      opq_entry_q_unlock(&opq);
      if (oqe->type == OQ_TXTMSG) {
         attribs[0] = INT2FIX(OQ_TXTMSG);
         attribs[1] = INT2NUM(oqe->ts);
         attribs[2] = rb_str_new2(oqe->msg);
         rb_yield(rb_class_new_instance(3, attribs, cSEmsg));

         free_stream_txt_object(
            (struct stream_txt *)oqe->msg, &stream_txt_space);
         if (strncmp(oqe->msg, "#Shut", 5) == 0)
	    exit(7);  /* Ruby code should have closed fdf */
         }
      else {
   	 sa[0] = INT2FIX(oqe->s_id);  /* Stream, only has @id so far */
         attribs[0] = INT2FIX(oqe->type);
	 attribs[1] = rb_class_new_instance(1, sa, cStream);
         attribs[2] = Qnil;
         rb_yield(rb_class_new_instance(3, attribs, cSEmsg));
         }
      free_opq_entry_object(oqe, &opq_entry_space);

      t_semaphore_up(&opq_empty);  /* Signal an empty block */
      }
   return self;
   }

static int get_stream_sizes(struct stream *sp) {
   rb_yield(INT2FIX(sp->tp));
   return 0;
   }

static VALUE owtm_each_stream_size(VALUE self) {
   /* CAUTION: locks stream table during walk */
   struct flow_watcher_control *fw;
   struct stream_queue *sq;
   int nfw, k;

   mutex_lock("ess", &stream_data_mutex);
   nfw = flow_watcher_control_space.mx_objects;
   for (k = 0; k != nfw; ++k) {
      fw = &flow_watcher_control_space.objects[k];
      sq = &fw->sq;
      stream_walk_up(sq, get_stream_sizes);
      }
   mutex_unlock("ess", &stream_data_mutex);
   return self;
   }

#if STREAMTIMEOUTS
int deleted_inactive, deleted_low_rate;  //dd
static void delete_stream(struct stream *sp) {
   sp->state = SS_DELETE;  /* Let watcher time it out */
   sp->timeout = STO_TWO_WAY;
   }
#endif

static int get_stream_attribs(struct stream *sp)
{  /* Return 0 to continue, 1 to stop walking */
   VALUE ra[13];  /* src_type  probe_port  enough_lifetime  to_pkts
     total_lifetime  mean_retry_time  n_dest_ports pkt_len_mode
     src_bin  tot_count  tcp_pkts udp_pkts other_pkts  */
  
#if STREAMTIMEOUTS
   if (sp->state != SS_NEW && sp->state != SS_ENOUGH) {
      two_way_streams += 1;
      two_way_pkts += (sp->tp + sp->fp);
      two_way_bytes += (sp->tb + sp->fb);
      if (sp->version == 6) {
         v6_two_way_streams += 1;
         v6_two_way_pkts += (sp->tp + sp->fp);
         v6_two_way_bytes += (sp->tb + sp->fb);
         }
      return 0;  /* Two-way, Delete or Timeout */
      }
   if (sp->active_since_last_sum == 0) {
      delete_stream(sp);  /* Inactive for a Record interval */
      deleted_inactive += 1;  //dd
      return 0;
      }
   else sp->active_since_last_sum = 0;
#endif

   one_way_streams += 1;
   one_way_pkts += sp->tp;  one_way_bytes += sp->tb;
   if (sp->version == 6) {
      v6_one_way_streams += 1;
      v6_one_way_pkts += sp->tp;  v6_one_way_bytes += sp->tb;
      }

   ra[0] = INT2FIX(sp->src_type);
   ra[1] = sp->probe_port_valid ? INT2FIX(sp->probe_port) : Qnil;
   ra[2] = rb_float_new(sp->enough_lifetime);
   ra[4] = rb_float_new(sp->last_us - sp->first_us);

   if (sp->analysed) {
      ra[5] = rb_float_new(sp->retry_time);
      ra[6] = INT2FIX(sp->n_dest_ports);
      ra[7] = INT2FIX(sp->pkt_len_mode);
      }
   else ra[5] = ra[6] = ra[7] = Qnil;
#if MKDESTCOUNTS
   ra[8] = INT2FIX(src_addr_bin(sp->src_addr.a.v4));
 #else
   ra[8] = Qnil
#endif
   ra[9] =  ULL2NUM(sp->totp);
   ra[10] = INT2NUM(sp->tcp_pkts);  ra[11] = INT2NUM(sp->udp_pkts);
   ra[12] = INT2NUM(sp->other_pkts);
   ra[3] = INT2FIX(sp->tp);
   sp->tcp_pkts = sp->udp_pkts = sp->other_pkts = sp->tp = 0;
   rb_yield(rb_ary_new4(sizeof(ra)/sizeof(VALUE), ra));

   return 0;
   }

static VALUE owtm_each_stream(VALUE self) {  /* pkt_datas for src_host self */
   struct flow_watcher_control *fw;
   struct stream_queue *sq;
   int nfw, k;
   struct stream_txt *st;

   nfw = flow_watcher_control_space.mx_objects;
#if STREAMTIMEOUTS
   for (;;) {  /* Wait for CV_IDLE */
      pthread_mutex_lock(&count_mutex);
      if (CV_IDLE) break;  /* Leave cv locked */
      else if (CV_CHECKING)  /* Must wait for timeouts to finish */
         pthread_cond_wait(&count_cv, &count_mutex);
      else log_msg(LOG_ERR, 1, "Attempt to Record: while CV_RECORDING <<<");
      pthread_mutex_unlock(&count_mutex);  /* Let watchers change count */
      }
   count -= 1;  /* Set count = CV_RECORDING */
   pthread_mutex_unlock(&count_mutex);
   deleted_inactive = deleted_low_rate = 0;  //dd
#endif
   for (k = 0; k != nfw; ++k) {  /* Record: processing */
      fw = &flow_watcher_control_space.objects[k];
      sq = &fw->sq;
      stream_q_lock(sq);
      stream_walk_up(sq, get_stream_attribs);
      stream_q_unlock(sq);
      }
#if STREAMTIMEOUTS
   pthread_mutex_lock(&count_mutex);
   count += 1;  /* Set count = CV_IDLE */
   pthread_mutex_unlock(&count_mutex);
   st = get_stream_txt_object(&stream_txt_space);
   sprintf(st->msg ,"#EachStream: inactive=%d, low_rate=%d",
      deleted_inactive, deleted_low_rate);
   enqueue_stream_txt(st->msg);
#endif
   return self;
   }

#if 0  //??
static int get_stream_info(struct stream *sp)
{  /* Return 0 to continue, 1 to stop walking */
      /* src_type  to_pkts  first_us last_us  probe_port  src_addr */
  VALUE ra[6], pa[3];
  uint8_t *a;

   ra[0] = INT2FIX(sp->src_type);
   ra[1] = INT2FIX(sp->tp);
   ra[2] = rb_float_new(sp->first_us);
   ra[3] = rb_float_new(sp->last_us);
   ra[4] = sp->probe_port_valid ? INT2FIX(sp->probe_port) : Qnil;
   if (sp->version == 4) {
     pa[0] = INT2FIX(4);  pa[2] = Qnil;
      a = sp->src_addr.a.a;
      pa[1] = rb_str_new((char *)a, IP4_ADDR_LEN);
      ra[5] = rb_class_new_instance(3, pa, IPprefix);
      }
   else {  /* v6 */
      pa[0] = INT2FIX(6);  pa[2] = Qnil;
      a = sp->src_addr.a.a;
      pa[1] = rb_str_new((char *)a, IP6_ADDR_LEN);
      ra[5] = rb_class_new_instance(3, pa, IPprefix);
      }

   rb_yield(rb_ary_new4(sizeof(ra)/sizeof(VALUE), ra));
   return 0;
   }

static VALUE owtm_each_str_info(VALUE self) {  /* for owt-sti.rb */
   struct flow_watcher_control *fw;
   struct stream_queue *sq;
   int nfw, k;

   nfw = flow_watcher_control_space.mx_objects;
   for (k = 0; k != nfw; ++k) {
      fw = &flow_watcher_control_space.objects[k];
      sq = &fw->sq;
      stream_walk_up(sq, get_stream_info);
      }
   return self;
   }
#endif

static VALUE owtm_interval_stats(VALUE self) {
#if MKSTATSCOUNTS
   VALUE ra[2];
   VALUE as[mx_source_types], uh[mx_source_types];
   //x      ps[mx_source_types];
   int st, m;
   for (st = 0; st != mx_source_types; ++st) {
      VALUE active_sources[STATS_INTERVALS_PER_HOUR+1],
         /* +1 because we see a few packets recorded after the hour change! */
         /*   Also, type 0 sources are recorded only at EOF */
	//x	 pkts_seen[STATS_INTERVALS_PER_HOUR+1],
	 unique_hosts[STATS_INTERVALS_PER_HOUR+1];
      for (m = 0; m != STATS_INTERVALS_PER_HOUR+1; ++m) {
        active_sources[m] = UINT2NUM(active_sources_per_stats_interval[m][st]);
        unique_hosts[m] = UINT2NUM(unique_hosts_per_stats_interval[m][st]);
	//x        pkts_seen[m] = UINT2NUM(pkts_seen_per_stats_interval[m][st]);
        }
      as[st] = rb_ary_new4(STATS_INTERVALS_PER_HOUR+1, active_sources);
      uh[st] = rb_ary_new4(STATS_INTERVALS_PER_HOUR+1, unique_hosts);
      //x      ps[st] = rb_ary_new4(STATS_INTERVALS_PER_HOUR+1, pkts_seen);
      }
   ra[0] = rb_ary_new4(mx_source_types, as);
   ra[1] = rb_ary_new4(mx_source_types, uh);
   //x   ra[2] = rb_ary_new4(mx_source_types, ps);
   return rb_ary_new4(2, ra);
#else
   return Qnil;
#endif
   }

static VALUE owtm_c_addr_distribs(VALUE self) {
   VALUE counts[13];
   counts[0] = ULL2NUM(one_way_pkts);  counts[1] = ULL2NUM(one_way_bytes);
   counts[2] = ULL2NUM(two_way_pkts);  counts[3] = ULL2NUM(two_way_bytes);
   counts[4] = ULL2NUM(one_way_streams);  counts[5] = ULL2NUM(two_way_streams);
   counts[6] = ULL2NUM(not_one_way_streams);
   counts[7] = ULL2NUM(v6_one_way_pkts);
   counts[8] = ULL2NUM(v6_one_way_bytes);
   counts[9] = ULL2NUM(v6_two_way_pkts);
   counts[10] = ULL2NUM(v6_two_way_bytes);
   counts[11] = ULL2NUM(v6_one_way_streams);
   counts[12] = ULL2NUM(v6_two_way_streams);
   VALUE ra[2];
   ra[0] = rb_ary_new4(sizeof(counts)/sizeof(VALUE), counts);
   ra[1] = Qnil;
#if MKDESTCOUNTS
   VALUE val[NDESTBINS];
   VALUE das[mx_source_types];
   int st, bin;
   for (st = 0; st != mx_source_types; ++st) {
      for (bin = 0; bin != NDESTBINS; ++bin) {
 	 val[bin] = UINT2NUM(dest_addrs_per_source_type[st][bin]);
         }
      das[st] = rb_ary_new4(NDESTBINS, val);
      }
   ra[1] = rb_ary_new4(mx_source_types, das);
#endif
   reset_summary_info();
   return rb_ary_new4(2, ra);
   }

#if 0  /* Gives Ruby GC problems when called from C thread! */
static VALUE glob_next_uri(VALUE self, VALUE fnv) {
   VALUE fn_rv =  /* global functions live in Kernel class */
      rb_funcall(rb_mKernel, rb_intern("next_uri"), 1, fnv);
   return fn_rv;
   }

char *c_next_uri(char *uri) {
   VALUE uri_v = rb_str_new2(uri);
   VALUE new_uri_v =  /* global functions live in Kernel class */
      rb_funcall(rb_mKernel, rb_intern("next_uri"), 1, uri_v);
   char *ustr = RSTRING(new_uri_v)->ptr;
   int ulen = RSTRING(new_uri_v)->len;
   char *rs = malloc(ulen+2);
   strcpy(rs, ustr);
   return rs;
   }
#endif

void Init_OWTmonitor() {
   /* rb_require("IPprefix");
         So that we can run in owt-cron-env environment, owtmon.rb
         must require IPprefix.so  before  require OWTmonitor */
   IPprefix = rb_path2class("IPprefix");

   mSE = rb_define_module("OWTmonitor");

   rb_define_global_const("OQ_TXTMSG",  INT2FIX(OQ_TXTMSG));
   rb_define_global_const("OQ_STARTED",  INT2FIX(OQ_STARTED));
   rb_define_global_const("OQ_PROGRESS", INT2FIX(OQ_PROGRESS));
   rb_define_global_const("OQ_ENOUGH", INT2FIX(OQ_ENOUGH));
   rb_define_global_const("OQ_TIMED_OUT", INT2FIX(OQ_TIMED_OUT));
   rb_define_global_const("OQ_END_OF_FILE", INT2FIX(OQ_END_OF_FILE));

   cStream = rb_define_class("OWTevent", rb_cObject);
   memset(&cs_oqe, 0, sizeof(struct opq_entry));
   rb_define_method(cStream, "initialize", cs_init, 1);
   rb_define_method(cStream, "last_tsec", cs_last_tsec, 0);
   rb_define_method(cStream, "usage", cs_usage, 0);
   rb_define_method(cStream, "each_pkt", cs_each_pkt, 0);
   rb_define_method(cStream, "set_source_attribs", cs_set_source_attribs, 1);

   cST = rb_define_class_under(mSE, "Monitor", rb_cObject);
   rb_define_method(cST, "initialize", owtm_initialize, 1);
   rb_define_method(cST, "owt_init", owtm_init, -1);
   rb_define_method(cST, "info", owtm_info, 0);
   rb_define_method(cST, "log_msg", owtm_log_msg, 2);
   rb_define_method(cST, "open", owtm_open, 0);
   rb_define_method(cST, "set_uris", owtm_set_uris, 1);
   rb_define_method(cST, "set_local_addrs", owtm_set_local_addrs, 1);
   rb_define_method(cST, "close", owtm_close, 0);
   rb_define_method(cST, "nodename", owtm_nodename, 0);
   rb_define_method(cST, "each", owtm_each, 0);
   rb_define_method(cST, "each_stream_size", owtm_each_stream_size, 0);
   rb_define_method(cST, "each_stream", owtm_each_stream, 0);
   //??   rb_define_method(cST, "each_str_info", owtm_each_str_info, 0);
   rb_define_method(cST, "interval_stats", owtm_interval_stats, 0);
   rb_define_method(cST, "c_addr_distribs", owtm_c_addr_distribs, 0);

   cSEmsg = rb_define_class("OWTmessage", rb_cObject);
   rb_define_method(cSEmsg, "initialize", msg_init, 3);
   }
