/* 1552, Thu 17 Jun 10 (NZST)
   1955, Thu 22 Apr 10 (PDT)
   1330, Fri  1 Feb 08 (NZDT)

   owt_watcher.c:  Packet and stream routines for owtmon

   Copyright 2007-2010 by Nevil Brownlee, U Auckland | CAIDA | WAND */

#include "owt.h"

/* Declarations for SRI code to generate Conficker port numbers (in conf-sri.c) */
typedef union {
   int16_t s16[8];  uint16_t u16[8];  int32_t s32[4];
   } result_t;
int portgen(int ip, result_t *res, int week);

uint32_t date_seed;

void set_date_seed(uint32_t ts) {
   /* Emile's python code ...
      date_seed = ((((uint64_t)ts - 345600) * 1861606989) >> 32) >> 18; */
   /* Nevil's C code: following line seems to be equivalent to
      date_seed = ((uint64_t)ts - 4*24*3600)/(7*24*3600); */
   date_seed = ((((uint64_t)ts - 345600) * 1861606989) >> 32) >> 18;
   //d printf("   l_t_s=%u, date_seed=%x\n", ts, date_seed);
   }

void swap_ip_addresses(struct ip_address *s, struct ip_address *d) {
   struct ip_address temp;
   memcpy(&temp, s, sizeof(struct ip_address));
   memcpy(s, d, sizeof(struct ip_address));
   memcpy(d, &temp, sizeof(struct ip_address));
   }

int get_v4key(struct stream **r_sp, int *r_fwd, int hl, unsigned char *hdr)
{  /* Code from pkt_extract() in NeTraMet's flowhash.c */
   struct stream *sp;
   struct pkt_data *pktp;
   int ver, dx, frag;
   struct ip_address *saddr, *daddr;  uint16_t sport, dport;
   int sa_local, da_local, fwd;

   if ((ver = hdr[0] >> 4) == 4) {
      if (hl < 12+IP4_ADDR_LEN*2) {
         *r_sp = NULL;
         return SKEY_NOPEERADDR;  /* Didn't get the peer addresses! */
         }
      pktp = (struct pkt_data *)get_pkt_data_object(&pkt_data_space);
      if (pktp == NULL) {
         *r_sp = NULL;
         return SKEY_NOSPACE;  /* No pkt_data space */
         }
      pktp->last_op = PD_READ;
      pktp->proto = hdr[9]; /* Protocol Type */

      if (frag = (ntohs(*(uint16_t *)&hdr[6]) & 0x1FFF) != 0) {
 	 free_pkt_data_object(pktp, &pkt_data_space);
         *r_sp = NULL;
         return SKEY_FRAGPKT;  /* Non-first fragment */
         }
      else {
         dx = (hdr[0] & 0x0F) << 2;  /* IPv4 header length in 32-bit units */
         pktp->tcp_flags = 0;
         switch (pktp->proto) {
         case PT_TCP:
	    pktp->tcp_flags = hl >= dx+14 ? hdr[dx+13] : 64;
         case PT_UDP:
         case PT_SCTP:
            if (hl < dx+4) {
   	       free_pkt_data_object(pktp, &pkt_data_space);
               *r_sp = NULL;
               return SKEY_NOTRANSADDR;  /* Didn't get the IP ports! */
               }
            else {
               sport = ntohs(*(uint16_t *)&hdr[dx]);  /* Source port */
               dport = ntohs(*(uint16_t *)&hdr[dx+2]);  /* Dest port */
	       }
            break;
         case PT_ICMP:
 	    pktp->tcp_flags = hdr[dx];  /* ICMP type */
         default:
            sport = dport = 0;
            break;
            }
         }
      //d      printf("get_v4key(1): pktp=%0x, saddr=%x, daddr=%x, sp=%x\n",
      //d	     pktp, saddr, daddr, sp);

      sp = get_stream(4);  sp->version = 4;
      saddr = &sp->src_addr;  daddr = &pktp->dst_addr;
      saddr->ver = daddr->ver = 4;
      saddr->a.v4 = *(uint32_t *)&hdr[12];  /* IPv4 source */
      daddr->a.v4 = *(uint32_t *)&hdr[16];  /* IPv4 dest */

      sa_local = addr_is_local(4, saddr);
      da_local = addr_is_local(4, daddr);
      if (sa_local && !da_local) fwd = 0;
      else if (!sa_local && da_local) fwd = 1;
      else {
         free_pkt_data_object(pktp, &pkt_data_space);
 	 free_stream(sp);
         *r_sp = NULL;
         return SKEY_NOLOCALADDR;
         }
      sp->pktp = pktp;

      if (fwd) {
         pktp->src_port = sport;  pktp->dst_port = dport;
         }
      else {
         pktp->src_port = dport;  pktp->dst_port = sport;
         swap_ip_addresses(saddr, daddr);
         }
      //      printf("*** src=%08x, dst=%08x, fwd=%d\n",  //p
      //	     sd4p->src_addr, sd4p->dst_addr, fwd);

      pktp->ip_len = ntohs(*(uint16_t *)&hdr[2]);

      pktp->cfc = 0;  /* SRI test for Conficker p2p source */
      /* Any UDP, any TCP not to port 445 or from port 80 */
      if (pktp->proto == PT_UDP || (pktp->proto == PT_TCP &&
            !(pktp->dst_port == 445 || pktp->src_port == 80)) ) {
	 int rc, i, j;	 result_t res;
	 rc = portgen(pktp->dst_addr.a.v4, &res, date_seed);
	 for (i = 0; i < 8; i++) {
	    if (res.u16[i] == pktp->dst_port) {
	       pktp->cfc = 1;
	       //d printf("get_v4key(): cfc, proto=%d, dst_addr=%x, dst_port=%d\n",
	       //d    pktp->proto, pktp->dst_addr->a.v4, pktp->dst_port);
               //d printf(".");
	       break;  /* Need only match one port! */
	       }
	    }
         }

      *r_fwd = fwd;
      *r_sp = sp;  /* Return stream (with ptrs to pkt_data and sd_tuple_v4) */
      return SKEY_OK;
      }
   else return SKEY_BADIPHDR;
   }

#define PT_6HOP       0  /* Hop-by-hop options */
#define PT_6ROUT     43  /* Routing (Type 0) */
#define PT_6FRAG     44  /* Fragment */
#define PT_6ESP      50  /* Encapsulating Security Payload (ESP) RFC 2406 */
#define PT_6AH       51  /* Authentication (AH) RFC 2402 */
#define PT_6ICMP     58  /* ICMPv6 */
#define PT_6NONH     59  /* No next header */
#define PT_6DEST     60  /* Destination options */
#define PT_6MOBILE  135  /* IPv6 Mobility Header */ 

int find_v6_type(int *prot, int *hx, const uint8_t *p, const int hl)
{
   int nh, nx;

   nh = p[6];  nx = 40;
   for (;;) {
      switch (nh) {
      case PT_6FRAG:    /* Fragment */
      case PT_6ESP:     /* Encapsulating Security Payload (ESP) RFC 2406 */
      case PT_6NONH:    /* No next header */
         *prot = nh;
         return 0;  /* Can't analyse this packet */

      case PT_6HOP:     /* Hop-by-hop options */
      case PT_6ROUT:    /* Routing (Type 0) */
      case PT_6DEST:    /* Destination options */
      case PT_6AH:      /* Authentication (AH) RFC 2402 */
      case PT_6MOBILE:  /* IPv6 Mobility Header */
	 break;  /* Keep looking */

      default:
         *prot = nh;  *hx = nx;
         return 1;  /* Can see into this one */
         }
      if (nx >= hl) {
	 *prot = -1;  /* Not enough header bytes to get Next Header */
         return 0;
         }
      nh = p[nx];
      nx += (p[nx+1]+1) * 8;
      }
   }

int get_v6key(struct stream **r_sp, int *r_fwd, int hl, unsigned char *hdr)
{  /* Code from pkt_extract() in NeTraMet's flowhash.c */
   struct stream *sp;
   struct pkt_data *pktp;
   struct ip_address *saddr, *daddr;  uint16_t sport, dport;
   int ver, dx, prot, c;
   int sa_local, da_local, fwd;

   if ((ver = hdr[0] >> 4) == 6) {
      if (hl < 8+IP6_ADDR_LEN*2) {
         *r_sp = NULL;
         return SKEY_NOPEERADDR;  /* Didn't get the peer addresses! */
         }
      pktp = (struct pkt_data *)get_pkt_data_object(&pkt_data_space);
      if (pktp == NULL) {
         *r_sp = NULL;
         return SKEY_NOSPACE;  /* No pkt_data space */
         }
      pktp->last_op = PD_READ;

      if (!find_v6_type(&prot,&dx, hdr,hl)) {
 	 free_pkt_data_object(pktp, &pkt_data_space);
         *r_sp = NULL;
         if (prot < 0)  /* Not enough bytes to determine protocol */
  	    return SKEY_V6SMALLSNAP;
         else return SKEY_V6OPAQUE;  /* Can't see into packet */
         }
      pktp->proto = prot;

      pktp->tcp_flags = 0;
      switch (prot) {
      case PT_TCP:
         pktp->tcp_flags = hl >= dx+14 ? hdr[dx+13] : 64;
      case PT_UDP:
      case PT_SCTP:
         if (hl < dx+4) {
  	    free_pkt_data_object(pktp, &pkt_data_space);
            *r_sp = NULL;
            return SKEY_NOTRANSADDR;  /* Didn't get the IP ports! */
            }
         else {
            sport = ntohs(*(uint16_t *)&hdr[dx]);  /* Source port */
            dport = ntohs(*(uint16_t *)&hdr[dx+2]);  /* Dest port */
            }
         break;
      case PT_6ICMP:
 	 pktp->tcp_flags = hdr[dx];  /* ICMP6 type */
      default:
         sport = dport = 0;
         break;
         }

      sp = get_stream(6);  sp->version = 6;
      saddr = &sp->src_addr;  daddr = &pktp->dst_addr;
      saddr->ver = daddr->ver = 6;
      memcpy(saddr->a.a, &hdr[8], IP6_ADDR_LEN);  /* IPv6 source */
      memcpy(daddr->a.a, &hdr[24],IP6_ADDR_LEN);  /* IPv6 dest */

      sa_local = addr_is_local(6, saddr);
      da_local = addr_is_local(6, daddr);
      if (sa_local && !da_local) fwd = 0;
      else if (!sa_local && da_local) fwd = 1;
      else {
         free_pkt_data_object(pktp, &pkt_data_space);
 	 free_stream(sp);
         *r_sp = NULL;
         return SKEY_NOLOCALADDR;
         }
      sp->pktp = pktp;

      if (fwd) {
         pktp->src_port = sport;  pktp->dst_port = dport;
         }
      else {
         pktp->src_port = dport;  pktp->dst_port = sport;
         swap_ip_addresses(saddr, daddr);
         }

      pktp->ip_len = 40 + ntohs(*(uint16_t *)&hdr[4]);
         /* Header + payload (includes extension headers) */

      *r_fwd = fwd;
      *r_sp = sp;  /* Return stream (with ptrs to pkt_data and sd_tuple_v6) */
      return SKEY_OK;
      }
   else return SKEY_BADIPHDR;  /* Wasn't a v6 packet! */
   }

static char key_s[150];  /* Return value for key_to_s() */
char *key_to_s(struct stream *sp) {

   if (sp->version == 4) {
      uint8_t *a = sp->src_addr.a.a;
      sprintf(key_s, "%u.%u.%u.%u", a[0],a[1],a[2],a[3]);
      }
   else if (sp->version == 6) {
      char *kp;
      kp = v6addr_to_s(key_s, &sp->src_addr);
      kp[1] = '\0';
      }
   else {
      quack(999);
      log_msg(LOG_ERR, 2,
          "Tried to print key_type %d <<<", sp->version);
      }
   return key_s;
   }

void print_stream(struct stream *sp) {
   if (sp->version != 4) printf("stream: type=%d <<<\n", sp->version);
   else {
      uint8_t *a = sp->src_addr.a.a;
      printf("%u.%u.%u.%u", a[0],a[1],a[2],a[3]);
      a = sp->pktp->dst_addr.a.a;
      printf("->%u.%u.%u.%u", a[0],a[1],a[2],a[3]);
      }
   }

void check_str_nulls(struct stream_queue *sq, struct stream *sp);

void enqueue_stream_status(int type, struct stream *sp, uint32_t now_ts) {
   struct opq_entry *oqe;
   t_semaphore_down(&opq_empty);  /* Wait for an empty entry */
   oqe = oqe_for_stream(sp);
   oqe->type = type;
   opq_entry_q_lock(&opq);
   opq_entry_push(&opq, oqe);
   opq_entry_q_unlock(&opq);
   sp->last_report_sec = now_ts;
   t_semaphore_up(&opq_ready);  /* Signal a full entry */
   }

void enqueue_stream_txt(char *msg) {
   struct opq_entry *oqe;
   t_semaphore_down(&opq_empty);  /* Wait for an empty entry */
   oqe = oqe_for_stream(NULL);
   oqe->type = OQ_TXTMSG;  oqe->msg = msg;
   oqe->ts = g_last_ts;
   opq_entry_q_lock(&opq);
   opq_entry_push(&opq, oqe);
   opq_entry_q_unlock(&opq);
   t_semaphore_up(&opq_ready);  /* Signal a full entry */
   }

void check_stream_timeouts(struct flow_watcher_control *fw, uint32_t now_ts) {
   struct stream_queue *sq = &fw->sq;
   double min_idle_us2, min_idle_us1, last_t_us, av_int;
   struct stream *sp;
   int keep, terminated, n, ow_terminated;
   int ignored = 0, counted = 0, freed = 0, send_stream_status;
   struct stream_data *sdp, *mdp;

#if SC_DEBUG >= 2
   fprintf(stderr, "  === check_timeouts(%d), ts=%u\n", fw->id, now_ts);
#endif
   last_t_us = now_ts*1000000.0;
   min_idle_us1 = (now_ts - ONE_WAY_TIMEOUT)*1000000.0;
   min_idle_us2 = (now_ts - TWO_WAY_TIMEOUT)*1000000.0;
   ow_terminated = terminated = 0;
#if SC_DEBUG >= 1
   check_stream_chains("--- starting check_stream_timeouts", sq);
#endif
   stream_q_lock(sq);
   for (n = sq->length; n != 0;  --n) {  /* Don't go round more than once */
      counted += 1;
      sp = stream_shift(sq);  /* Remove stream from head */
      if (sp->last_us > (sp->timeout == STO_ONE_WAY ?
            min_idle_us1 : min_idle_us2) ) {
 	 stream_unshift(sq, sp);  /* Put it back! */
         break;  /* Hasn't been queued long enough; stop checking */
         }
      else {  /* Could be inactive, can we recover it? */
         if (sp->totp == 1)
            keep = 0;  /* Only one packet seen */
         else {
	    av_int = (sp->last_us - sp->first_us)/
               (double)(sp->totp - 1);
            keep = (last_t_us - sp->last_us)  /* Observed idle us */
	       < av_int*STO_MULTIPLIER;  /* Current timeout estimate */
            }
         }
      if (keep) stream_push(sq, sp);  /* Add stream to tail */
         /* We won't waste time tesing it until it drifts back
            to the head of the queue.  That means we no longer
            have the queue in (approximnate) LastTime order */
      else {  /* Timed-out stream */
         terminated += 1;
         if (sp->state == SS_NEW) {
            if (sp->totp > sp->pkts_last_classify) {
  	       if ((ow_terminated += 1) > MX_ONE_WAY_TIMEOUTS_PER_CHECK) {
   	          stream_unshift(sq, sp);  /* Put it back at head */
                  terminated -= 1;
                  break;  /* Limit opq load */
	          }
               sp->enough_lifetime = 0;  /* Might time out more than once! */
               /* Try to classify it (better) now */
	       if (sp->msg_queued == OQ_NULL) {
 	          enqueue_stream_status(OQ_TIMED_OUT, sp, now_ts);
                  sp->msg_queued = OQ_TIMED_OUT;
	          }
	       sp->pkts_last_classify = sp->totp;  /* Don't queue it again! */
	       }
            stream_push(sq, sp);  /* Put it back at tail */
	    }
         else if (sp->state == SS_ENOUGH) {
            stream_push(sq, sp);  /* Put it back at tail */
	    }
	 else {  /* Two-way */
            sp->state = SS_TIMEOUT;
	    two_way_streams += 1;
	    two_way_pkts += (sp->tp + sp->fp);
	    two_way_bytes += (sp->tb + sp->fb);
            if (sp->version == 6) {
               v6_two_way_streams += 1;
               v6_two_way_pkts += (sp->tp + sp->fp);
               v6_two_way_bytes += (sp->tb + sp->fb);
               }
            owt_ht_delete(fw->sh_table, sp);  /* Remove from hash table */
	    sp->next = sp->prev = (struct stream *)0x55555555;
            free_stream(sp);
	    freed += 1;
            }
	 //	 check_stream_chains("--- after timed out", sq);
         }
      }
#if SC_DEBUG >= 2
n = stream_space.mx_objects - stream_space.free_queue.length;
   fprintf(stderr, "   ==== %d streams terminated, %d freed, %d ignored, %d counted.  %d active sv4s\n",
	   terminated, freed, ignored, counted, n);
#endif

   stream_q_unlock(sq);
   }

uint32_t swp_now_ts;
int get_stream_with_pkts(struct stream *sp)
{  /* Return 0 to continue, 1 to stop walking */
   if (sp->state == SS_NEW) {  /* One-way stream */
      /* Looking at every packet doesn't seem to add much elapsed time */
      sp->enough_lifetime = sp->last_us - sp->first_us;
      enqueue_stream_status(OQ_END_OF_FILE, sp, swp_now_ts);
      }
   return 0;
   }

void output_active_streams(struct stream_queue *sq, uint32_t now_ts) {
   stream_q_lock(sq);
   check_stream_chains("--- starting output_active_src_hosts",sq);
   swp_now_ts = now_ts;
   stream_walk_up(sq, get_stream_with_pkts);
   check_stream_chains("+++ leaving output_active_src_hosts", sq);
   stream_q_unlock(sq);
   }

void check_str_nulls(struct stream_queue *sq, struct stream *sp) {
   if (sp->next == NULL && sp != sq->tail) {
      printf("null sp.next and sp != sp tail ***\n");
      quack(991);
      }
   if (sp->prev == NULL && sp != sq->head) {
      printf("null sp.prev and sp != s head ***\n");
      quack(992);
      }
   }

static char ftb[32];
static char *fmt_time(int utc_time, uint32_t *ti) {  /* From NeTraMet */
#if SIZEOF_TIME_T == 8
   uint64_t lti = *ti;
#else
   uint32_t lti = *ti;
#endif
   if (!utc_time) {
      char *ts = ctime((time_t *)&lti);
      strncpy(&ftb[0],  &ts[11], 9);  /* 17:31:42_ */
      strncpy(&ftb[9],  &ts[0],  4);  /* Thu_ */
      strncpy(&ftb[13], &ts[8],  3);  /* 23_  */
      strncpy(&ftb[16], &ts[4],  4);  /* Sep_ */
      strncpy(&ftb[20], &ts[20], 4);  /* 1997 */
      ftb[24] = '\0';
      }
   else {
      struct tm *t = gmtime((time_t *)&lti);
      char *dow[] = {
         "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
      char *moy[] = {
         "Jan", "Feb", "Mar", "Apr", "May", "Jun",
         "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
      sprintf(ftb, "%02d:%02d:%02d %s %2d %s %4d",
         t->tm_hour,t->tm_min,t->tm_sec, dow[t->tm_wday],
	 t->tm_mday, moy[t->tm_mon], t->tm_year+1900);
      }
   return ftb;
   }   

static char *trace_name;
static uint32_t last_timerec_sec;
static int n_times = STOP_AFTER;
static void output_time_record(uint32_t now_ts) {
   struct stream_txt *st = get_stream_txt_object(&stream_txt_space);
   int j;  struct stream *mp;
   if (st != NULL) {
      sprintf(st->msg, "#Time: %s  %s  Flows from %lu to %lu",
         fmt_time(1, &now_ts), trace_name,
         (unsigned long)last_timerec_sec, (unsigned long)now_ts);
      enqueue_stream_txt(st->msg);
      }
   last_timerec_sec = now_ts;
#if STOP_AFTER
   if (--n_times == 0) request_shutdown = 1;
#endif
   }

static void fdf_hdr_string(char *buf) {
   sprintf(buf, "##%s: -c%u %s -s%u -p%u starting at %s",
      version, time_rec_interval, trace_name,
      stream_space.mx_objects,
      pkt_data_space.mx_objects, fmt_time(1, &start_ts));
   }

static void output_fdf_hdr(void) {
   struct stream_txt *st = get_stream_txt_object(&stream_txt_space);
   if (st != NULL) {
      fdf_hdr_string(st->msg);
      enqueue_stream_txt(st->msg);
      }
   }

char *owt_file_header(void) {
   char *msg = malloc(MXSTRMTXTLEN);
   fdf_hdr_string(msg);
   return msg;
   }

#define MONITOR_PKTD_STATE  0
#if MONITOR_PKTD_STATE
int nstreams[SS_n_states], npkts[SS_n_states];
int count_pkts(struct stream *sp) {
   nstreams[sp->state] += 1;
   npkts[sp->state] += sp->pktq.length;
   return 0;  /* keep walking */
   }

void gather_stream_pkt_stats(void) {
   int n_pkts[PD_n_states+1];
   struct flow_watcher_control *fw;
   int j;
   for (j = 0; j != SS_n_states;  ++j)
      nstreams[j] = npkts[j] = 0;
   fw =&flow_watcher_control_space.objects[0];
   stream_q_lock(&fw->sq);
   stream_walk_up(&fw->sq, count_pkts);
   stream_q_unlock(&fw->sq);
   printf("*** stream:pkts ");
   for (j = 0; j != SS_n_states;  ++j)
      printf("  %d %d %d", j, nstreams[j], npkts[j]);
   printf("\n");  fflush(stdout);

   for (j = 0; j != PD_n_states;  ++j) n_pkts[j] = 0;
   for (j = 0;  j != pkt_data_space.mx_objects; ++j)
      n_pkts[pkt_data_space.objects[j].last_op] += 1;
   printf("--- pkt_data:pkts ");
   for (j = 0; j != PD_n_states;  ++j)
      printf("  %d %d", j, n_pkts[j]);
   printf("\n");  fflush(stdout);
   }
#endif

void update_interface_stats(uint32_t now_ts) {
   /* Called every STATS_INTERVAL */
   int pd = pkt_drops();
   struct stream_txt *st = get_stream_txt_object(&stream_txt_space);
   char *stats_format = "#Stats: %u " \
         " aps=%u mps=%u lpk=%u lpp=%.3f  mst=%u mpd=%u acl=%.3f" \
         "  nip=%u nfr=%u nts=%u  mql=%u";

   if (st != NULL) {
      int j;
      uint32_t av_pkt_count, n_seconds, mx_pkt_rate;
      struct if_info *iip;

      av_pkt_count = n_seconds = mx_pkt_rate = 0;
      mutex_lock("uis", &if_info_mutex);
      for (j = 0; j != n_interfaces; ++j) {
         iip = &if_info[j];
         av_pkt_count += iip->av_pkt_count;
         if (iip->n_seconds > n_seconds) n_seconds = iip->n_seconds;
         if (iip->mx_pkt_rate > mx_pkt_rate) mx_pkt_rate = iip->mx_pkt_rate;
         iip->mx_pkt_rate = iip->av_pkt_count = iip->n_seconds = 0;
         }
      mutex_unlock("uis", &if_info_mutex);

      double apc = av_pkt_count == 0 ? 1.0 : av_pkt_count;
      double r_hc_searches = n_hc_searches == 0 ? 1.0 : (double)n_hc_searches;
      double acl = (double)av_cl_count/r_hc_searches;
      uint32_t aps = (n_seconds == 0) ? av_pkt_count
         : (av_pkt_count*10+5)/(n_seconds*10);
      sprintf(st->msg, stats_format, now_ts,
	 aps, mx_pkt_rate, pd, (double)pd*100.0/apc,
         mx_active_streams, mx_active_pkt_datas,
         acl, n_not_ip, n_frags, n_too_small,
         mx_active_opq_entries);
      enqueue_stream_txt(st->msg);
      }
   n_not_ip = n_frags = n_too_small = n_no_local_address =
      av_cl_count = n_hc_searches = n_v6_opaque =
      mx_active_streams = mx_active_pkt_datas = mx_active_opq_entries = 0;

#if MKSTATSCOUNTS  /* Move on to next interval */
   active_sources_this_interval = 
      active_sources_per_stats_interval[stats_interval_nbr];
   unique_hosts_this_interval =
      unique_hosts_per_stats_interval[stats_interval_nbr];
   //x   pkts_seen_this_interval = 
   //x      pkts_seen_per_stats_interval[stats_interval_nbr];
   ++stats_interval_nbr;
#endif

#if MONITOR_PKTD_STATE
   gather_stream_pkt_stats();
#endif
   }

static void each_time_rec(uint32_t now_ts) {
   /* Called every TIME_INTERVAL */
   output_time_record(now_ts);
#if STREAMTIMEOUTS  /* For trace file, just write summary at EOF */
   struct stream_txt *st = get_stream_txt_object(&stream_txt_space);
   if (st == NULL)
      log_msg(LOG_ERR, 1, "each_time_rec() couldn't get stream_txt");
   sprintf(st->msg, "#Record: time to write a sum file!");
   enqueue_stream_txt(st->msg);
#endif
   }

int fw_id = 0;
void start_flow_watcher(struct flow_watcher_control *fw, 
      int pw_q_sz, int ht_size) {
   int j;
   fw->id = fw_id;  fw_id += 1;
#if STARTUP_DEBUG
      printf("start_flow_watcher(1): pw_q_sz=%d, ht_size=%d\n",  //dd
   	  pw_q_sz, ht_size);  //dd
#endif
   fw->sh_table = owt_ht_create(ht_size);  /* Stream hash table */
   if (fw->sh_table == NULL)
      log_msg(LOG_ERR, 1, "Couldn't create hash table <<<");

   init_stream_queue(&fw->ispq);
   init_stream_queue(&fw->sq);
   t_semaphore_init(&fw->ispq_ready);
   t_semaphore_decrement(&fw->ispq_ready);  /* Thread blocked (v=0) */
   t_semaphore_init(&fw->ispq_empty);
   for (j = 0; j != pw_q_sz-1; ++j)
      t_semaphore_up(&fw->ispq_empty);  /* v=(nbr entries avail) */

   mutex_lock("sfw", &n_fw_mutex);
   n_fw_running += 1;
   mutex_unlock("sfw", &n_fw_mutex);

   fw->eof_sp = get_stream(6);
#if STARTUP_DEBUG
   printf("start_fw() about to call pthread_create()\n"); fflush(stdout);
#endif
   if (pthread_create(&fw->flow_thread, NULL, flow_watcher_process, fw)) {
      fprintf(stderr, "Error creating flow-watch thread <<<\n");
      exit(41);
      }
#if STARTUP_DEBUG
   printf("leaving start_fw +-+\n"); fflush(stdout);
#endif
   }

int n_flow_watchers, fw_next_st_to_check;

void reset_summary_info(void) {
   int j;
   one_way_bytes= two_way_bytes = v6_one_way_bytes = v6_two_way_bytes =
      one_way_pkts = two_way_pkts = v6_one_way_pkts = v6_two_way_pkts =
      one_way_streams = two_way_streams = not_one_way_streams =
      v6_one_way_streams = v6_two_way_streams = 0;
#if MKDESTCOUNTS
   for (j = 0; j != n_source_types; ++j) {
      memset(dest_addrs_per_source_type[j], 0, NDESTBINS*sizeof(uint32_t));
      }
#endif
   }

void start_meter(char *interfaces[], int n_ifs,
      int mx_events, int mx_opq_entries, 
      int mx_pkt_datas, int mx_streams,
      int pw_q_sz, int mx_stream_txts, int mx_flow_watchers,
      int mx_source_types, int hours) {
   struct event *ep;
   char name[250], *np;
   int err, r, j;  uint32_t ht_size;
   struct if_info *iip;
#if STARTUP_DEBUG
   printf("entering start_meter() +-+\n");  fflush(stdout);
#endif
   sprintf(name, "One-Way Traffic Monitor v%s", METER_VERSION);
   version = malloc(strlen(name)+1);
   strcpy(version, name);
   n_source_types = mx_source_types;

#if MKSTATSCOUNTS
#define N_INTERVALS  (traces_to_process*STATS_INTERVALS_PER_HOUR +1)
   active_sources_per_stats_interval =  /* Allocate per-src_type rows */
      calloc(N_INTERVALS, sizeof(uint32_t *));
      /* +1 because we see a few packets recorded after the hour change! */
   unique_hosts_per_stats_interval =
      calloc(N_INTERVALS,  sizeof(uint32_t *));
   //x   pkts_seen_per_stats_interval =
   //x      calloc(N_INTERVALS, sizeof(uint32_t *));
   for (j = 0; j != N_INTERVALS; ++j) {
      /* Allocate per-minute rows */
      active_sources_per_stats_interval[j] =
	calloc(n_source_types, sizeof(uint32_t));
      unique_hosts_per_stats_interval[j] =
	calloc(n_source_types, sizeof(uint32_t *));
      //x      pkts_seen_per_stats_interval[j] =
      //x	calloc(n_source_types, sizeof(uint32_t *));
      }
   active_sources_this_interval = active_sources_per_stats_interval[0];
   unique_hosts_this_interval = unique_hosts_per_stats_interval[0];
   //x   pkts_seen_this_interval = pkts_seen_per_stats_interval[0];
   stats_interval_nbr = 1;  /* 1-org interval nbrs */
#endif
#if MKDESTCOUNTS
   dest_addrs_per_source_type = malloc(n_source_types*sizeof(uint32_t *));
   for (j = 0; j != n_source_types; ++j) {
      dest_addrs_per_source_type[j] = calloc(NDESTBINS, sizeof(uint32_t));
      }
#endif

   request_shutdown = count = 0;
   owt_start(mx_events, mx_opq_entries,
      mx_streams, mx_pkt_datas,
      mx_stream_txts, mx_flow_watchers);
   last_timerec_sec = 0;  uname(&node);

   mx_active_streams = mx_active_pkt_datas = mx_active_opq_entries = 0;
   n_interfaces = n_ifs;
   for (err = j = 0; j != n_interfaces; ++j) {
      iip = &if_info[j];
      iip->if_name = interfaces[j];
      if ((r = init_trace(iip)) != 0)  /* Do this before scheduling events! */
	 err = 1;
      iip->if_nbr = j;  iip->n_pkts = 0;
      iip->mx_pkt_rate = iip->av_pkt_count = iip->n_seconds = 0;
      }
   if (err) exit(99);
   n_not_ip = n_frags = n_too_small = n_v6_opaque =
      intervals_this_sum_file = 0;

   sprintf(name, "%s", interfaces[0]);
   np = name + strlen(name);
   for (j = 1; j < n_ifs; ++j) {
      sprintf(np, ", %s", interfaces[j]);
      np = name + strlen(name);
      }
   sprintf(np, " on %s", node.nodename);
   trace_name = malloc(strlen(name)+1);
   strcpy(trace_name, name);

   t_semaphore_init(&opq_ready);
   t_semaphore_decrement(&opq_ready);  /* Thread blocked (v=0) */
   t_semaphore_init(&opq_empty);
   for (j = 0; j != opq_entry_space.mx_objects-1; ++j)
       t_semaphore_up(&opq_empty);  /* v=(nbr entries avail) */

   output_fdf_hdr();  /* Output .fdf header record(s) */

   ep = get_event(S_TIME_EVENT, 1);    /* Every 1s */ 
   schedule_event(ep, start_ts+3);  /* First time 3s from now */

   ep = get_event(STATS_EVENT, STATS_RECORDING_INTERVAL);
   schedule_event(ep, start_ts+4);

   ep = get_event(TIME_EVENT, time_rec_interval);
   schedule_event(ep, (start_ts+time_rec_interval-1)
      /time_rec_interval*time_rec_interval);  // 12 Apr 10, was +1

   n_flow_watchers = flow_watcher_control_space.mx_objects;
   n_fw_running = 0;
   fw_next_st_to_check = n_flow_watchers;
   one_way_bytes= two_way_bytes = v6_one_way_bytes = v6_two_way_bytes =
      one_way_pkts = two_way_pkts = v6_one_way_pkts = v6_two_way_pkts =
      one_way_streams = two_way_streams = not_one_way_streams =
      v6_one_way_streams = v6_two_way_streams = 0;
   last_drops = 0;

#if STARTUP_DEBUG
   printf("about to start flow-watch threads +-+\n");  fflush(stdout);
#endif
   /* Start flow-watching thread(s) */
   ht_size = stream_space.mx_objects*2/mx_flow_watchers;
   if (ht_size < MIN_HT_SIZE) ht_size = MIN_HT_SIZE;
#if STARTUP_DEBUG
      printf("start_meter(2.5):mx_stream=%u, mx_fw=%u, ht_size=%u\n",
         stream_space.mx_objects,mx_flow_watchers, ht_size);  //d
#endif
   for (j = 0; j != flow_watcher_control_space.mx_objects; ++j)
      start_flow_watcher(&flow_watcher_control_space.objects[j],
         pw_q_sz, ht_size);

#if STARTUP_DEBUG
   printf("about to start packet-wat threads +-+\n");  fflush(stdout);
#endif
   /* Start the packet-watching thread(s) */
   for (n_interfaces_running = j = 0; j != n_interfaces; ++j) {
      iip = &if_info[j];
      if (pthread_create(&iip->pw_thread, NULL, packet_watcher_process, iip)) {
         fprintf(stderr, "Error creating packet-watch thread <<<\n");
         exit(31);
         }
      iip->thread_running = 1;  n_interfaces_running += 1;
      }

   setgid(ut_gid);  setuid(ut_uid);  /* Don't need to run as root now */
   my_uid = ut_uid;  my_gid = ut_gid;

   log_msg(LOG_WARNING, 0, "Starting %s, running on %s",
      version, node.nodename);

#if STARTUP_DEBUG
   printf("leaving start_meter() +-+\n");  fflush(stdout);
#endif
   }

void stop_meter(char *msg) {
   struct stream_txt *st;
   int j;
   st = get_stream_txt_object(&stream_txt_space);
   strcpy(st->msg, msg);
   enqueue_stream_txt(st->msg);  /* Queue shutdown message */
   }

int check_stream_chains(char *msg, struct stream_queue *sq)
{  /* Consistency check for stream queue */
   struct stream *fsp,*rsp, *l_fsp, *l_rsp;
   char buf[100];
   int k, nas, rv;

   nas = sq->length;
   fsp = sq->head;  rsp = sq->tail;
   l_fsp = l_rsp = NULL;

   for (rv = 1, k = 0; ; ++k) {  /* Walk forward and reverse chains */
      if (fsp == NULL || rsp == NULL) {  /* Both NULL when done */
         if (fsp != NULL || fsp != rsp) {
	    log_msg(LOG_WARNING, 0,
               "s_c_c(%s): fsp=%x != rsp=%x, k=%d. nas=%d\n",
		    msg, fsp, rsp, k, nas);
 quack(419);
            rv = 0;
            }
         break;
         }
      if (k > nas+10) {
         log_msg(LOG_WARNING, 0,
            "s_c_c(%s): loop in chain, k=%d, nas=%d", msg, k, nas);
         fflush(stdout);
 quack(420);
         rv = 0;
         break;
         }
      l_fsp = fsp;  l_rsp = rsp;  /* For gdb debugging */
      fsp = fsp->next;  rsp = rsp->prev;
      }

   if (k != nas) {
      log_msg(LOG_WARNING, 0,
         "s_c_c(%s): k=%u != nas=%u", msg, k, nas);
 quack(421);
      printf("\n");
      rv = 0;
      }
   return rv;
   }

double at_seconds(struct timeval tv) {  /* Arrival time in seconds */
   return (double)(tv.tv_sec-start_ts) + (double)tv.tv_usec/1000000.0;
   }

uint32_t pkt_drops(void) {
   int j;  uint64_t r, drops = 0;
   for (j = 0; j != n_interfaces; ++j)
      drops += trace_get_dropped_packets(if_info[j].trace);
   r = drops - last_drops;
   last_drops = drops;
   return (uint32_t)r;
   }

int init_trace(struct if_info *iip) {
   struct timeval tv;

   iip->trace = NULL;  iip->packet = NULL;
   iip->trace = trace_create(iip->if_name);
   if (trace_is_err(iip->trace)) {
      log_msg(LOG_ERR, 0, "Couldn't open trace %s <<<", iip->if_name);
      return 101;
      }

   snaplen = SNAP_LEN;
   if (trace_config(iip->trace,TRACE_OPTION_SNAPLEN,&snaplen)) {
      log_msg(LOG_ERR, 0, "Couldn't set snap length %d <<<", snaplen);
      return 102;
      }

   if (trace_start(iip->trace)) {
      log_msg(LOG_ERR, 0, "Couldn't start trace %s <<<", iip->if_name);
      trace_destroy(iip->trace);
      return 103;
      }

   iip->packet = trace_create_packet();
   if (trace_read_packet(iip->trace, iip->packet) > 0) {
      if (start_ts == 0) {
         tv = trace_get_timeval(iip->packet); 
         start_ts = tv.tv_sec;
         set_date_seed(start_ts);  /* For Conficker p2p test */
         }
      }
   else {
      log_msg(LOG_ERR, 0, "Couldn't read packet from %s <<<", iip->if_name);
      return 104;
      }

   return 0;
   }

void close_trace(struct if_info *iip) {
   if (iip->trace != NULL) {
      if (trace_is_err(iip->trace)) log_msg(LOG_ERR, 1,
         "Error reading packets from %s !!!", iip->if_name);
      trace_destroy(iip->trace);
      }
   if (iip->packet != NULL)
      trace_destroy_packet(iip->packet);
   }

int next_trace_file(struct if_info *iip) {
   free(iip->if_name);  /* processewd uri */
   iip->if_name = other_uris[next_uri_nbr];
   next_uri_nbr += 1;

   iip->trace = NULL;  iip->packet = NULL;
   iip->trace = trace_create(iip->if_name);
   if (trace_is_err(iip->trace)) {
      log_msg(LOG_ERR, 0, "Couldn't open trace %s <<<", iip->if_name);
      return 101;
      }

   if (trace_config(iip->trace,TRACE_OPTION_SNAPLEN,&snaplen)) {
      log_msg(LOG_ERR, 0, "Couldn't set snap length %d <<<", snaplen);
      return 102;
      }

   if (trace_start(iip->trace)) {
      log_msg(LOG_ERR, 0, "Couldn't start trace %s <<<", iip->if_name);
      trace_destroy(iip->trace);
      return 103;
      }

   iip->packet = trace_create_packet();
   return 0;
   }

void one_second_process(uint32_t ts) {
   int j;
   struct if_info *iip;
   uint32_t active_ip_addresses,
      active_streams, active_pkt_datas, active_opq_entries;

   set_date_seed(ts);  /* For Conficker p2p test */

   for (j = 0; j != n_interfaces; ++j) {
      iip = &if_info[j];
      iip->av_pkt_count += iip->n_pkts;  iip->n_seconds += 1;
      if (iip->n_pkts > iip->mx_pkt_rate) iip->mx_pkt_rate = iip->n_pkts;
      iip->n_pkts = 0;  g_last_ts = iip->last_ts = ts;
      }
   mutex_unlock("osp", &if_info_mutex);  /* Locked in seconds_change() */

   active_streams = stream_space.mx_objects -
      stream_space.free_queue.length;
   if (active_streams > mx_active_streams)
      mx_active_streams = active_streams;
   active_pkt_datas = pkt_data_space.mx_objects -
      pkt_data_space.free_queue.length;
   if (active_pkt_datas > mx_active_pkt_datas)
      mx_active_pkt_datas = active_pkt_datas;
   active_opq_entries = opq_entry_space.mx_objects -
      opq_entry_space.free_queue.length;
   if (active_opq_entries > mx_active_opq_entries)
      mx_active_opq_entries = active_opq_entries;
   }

void seconds_change(struct if_info *iip, int ts) {
   uint32_t e_type;
   struct stream *event_sp;
   struct flow_watcher_control *fw;

   mutex_lock("osp", &if_info_mutex);
   if (ts <= g_last_ts) {
      iip->last_ts = ts;
      mutex_unlock("osp", &if_info_mutex);
      }
   else {  /* Change in overall last_ts */
      one_second_process(ts);  /* Update g_last_ts, unlock if_info */

      for (;;) {  /* Handle events */
         e_type = check_events(ts);
         if (e_type == NO_EVENT) break;
	 else switch (e_type) {
	 case S_TIME_EVENT:  /* Timeout streams */
	    event_sp = get_stream(6);

   	    fw_next_st_to_check -= 1;
	    fw = &flow_watcher_control_space.objects[
               fw_next_st_to_check];
            if (fw_next_st_to_check == 0)
               fw_next_st_to_check = n_flow_watchers;

	    t_semaphore_down(&fw->ispq_empty);
            event_sp->event_type = S_TIME_EVENT;
            event_sp->ts = g_last_ts;

            stream_q_lock(&fw->ispq);
            stream_push(&fw->ispq, event_sp);
            stream_q_unlock(&fw->ispq);
            t_semaphore_up(&fw->ispq_ready);
               /* Pass event to flow watcher */
            break;
	 case STATS_EVENT:  /* Write #Stats */
	    update_interface_stats(ts);
            break;
         case TIME_EVENT:  /* Write #Time */
	    each_time_rec(ts);
            break;
            }
         }  /* event handling loop */
      }
   }

void each_pkt(struct flow_watcher_control *fw, struct stream *isp) {
   struct stream *slr, *tq;
   int sec_since_last_pkt;
   int new = 1, src_new = 1, send_stream_status;

   slr = owt_ht_lookup(fw->sh_table, isp->hash, isp, &new);
                                         /* new => OK_to_insert */
   if (new) {  /* New stream, slr == isp */
      slr->state = SS_NEW;
      slr->version = isp->version;
      slr->first_fwd = isp->fwd;
      slr->tcp_pkts = slr->udp_pkts = slr->other_pkts =
         slr->tb = slr->fb = slr->tp = slr->fp = slr->totp = 0;
      slr->pkts_last_classify = 0;
      slr->first_us = slr->last_us;
      slr->last_report_sec = slr->ts;
      slr->timeout = STO_ONE_WAY;
      slr->src_addr = isp->src_addr;
      slr->msg_queued = OQ_NULL;
      init_pkt_data_queue(&slr->pktq);
      }
   else {  /* Already in table, copy fields from isp */
      if (slr == isp) quack(3004);
#if CHECKPKTQUEUE
      pkt_data_queue_ok(&slr->pktq, 1, 80);  //dd
#endif
      slr->last_us = isp->last_us;
      }

   if (isp->fwd) {
      mutex_lock("epk", &stream_data_mutex);
      slr->tp += 1;  slr->tb += isp->pktp->ip_len;
      mutex_unlock("epk", &stream_data_mutex);
      }
   else {
      mutex_lock("epk", &stream_data_mutex);
      slr->fp += 1;  slr->fb += isp->pktp->ip_len;
      mutex_unlock("epk", &stream_data_mutex);
      }
   slr->totp += 1;
   slr->active_since_last_sum = 1;
   if (isp->pktp->proto == PT_TCP) slr->tcp_pkts += 1;
   else if (isp->pktp->proto == PT_UDP) slr->udp_pkts += 1;
   else slr->other_pkts += 1;

#if CHECKPKTQUEUE
   pkt_data_queue_ok(&slr->pktq, 1, 82);  //dd
#endif

   isp->pktp->last_op = slr->state;  /* Save the state of it's stream */

   if (slr->state == SS_NEW) {  /* Not yet SS_TWO_WAY or SS_ENOUGH */
      if (slr->fp != 0) {  /* Pkts out from local net -> Two-way stream */
#if CHECKPKTQUEUE
         pkt_data_queue_ok(&slr->pktq, 0, 84);  //dd
#endif
         free_pkt_queue(&slr->pktq);  /* Free packets for two-way stream */
#if CHECKPKTQUEUE
         pkt_data_queue_ok(&slr->pktq, 0, 86);  //dd
#endif
         free_pkt_data(isp->pktp);  /* Discard the incoming pkt_data */
         slr->state = SS_TWO_WAY;
         slr->timeout = STO_TWO_WAY;
         }
      else {  /* Still one-way, add pkt to pktq */
#if CHECKPKTQUEUE
         pkt_data_queue_ok(&slr->pktq, 1, 88);  //dd
#endif
         isp->pktp->arrival_time = slr->last_us;
	 if (slr->tp <= pkts_to_record) {
            pkt_data_q_lock(&slr->pktq);
            pkt_data_push(&slr->pktq, isp->pktp);
            pkt_data_q_unlock(&slr->pktq);
	    if ((slr->tp) == pkts_to_record) {
	       slr->state = SS_ENOUGH;  /* No longer SS_NEW */
               slr->enough_lifetime = slr->last_us - slr->first_us;
               enqueue_stream_status(OQ_ENOUGH, slr, slr->ts);
	       }
	    }
#if CHECKPKTQUEUE
         pkt_data_queue_ok(&slr->pktq, 1, 96);
#endif
         }
      }
   else if (slr->state == SS_ENOUGH) {
#if CHECKPKTQUEUE
      pkt_data_queue_ok(&slr->pktq, 1, 98);
#endif
      if (slr->fp != 0) {  /* Two-way after all, may not be analysed yet! */
         slr->state = SS_TWO_WAY;
         slr->timeout = STO_TWO_WAY;
	 not_one_way_streams += 1;
         free_pkt_data(isp->pktp);  /* Discard the incoming pkt_data */
         }
      else {
#if MKSTATSCOUNTS 
         if (slr->analysed) {  /* src_type may not have been set yet */
	   //x         pkts_seen_this_interval[slr->src_type] += 1;
            if (slr->last_active_interval < stats_interval_nbr) {
               slr->last_active_interval = stats_interval_nbr;
               active_sources_this_interval[slr->src_type] += 1;
	       }
            }
#endif
#if MKDESTOUNTS 
         if (slr->analysed) {  /* src_type may not have been set yet */
            if (pktp->dst_addr.ver == 4) { 
               dest_addrs_per_source_type[sp->src_type][
                  dest_addr_bin(pktp->dst_addr.a.v4)] += 1; 
   	       } 
            }
#endif
         free_pkt_data(isp->pktp);  /* Discard the incoming pkt_data */
         }
      }
   else if (slr->state == SS_DELETE) {
      slr->state = SS_ENOUGH;
      free_pkt_data(isp->pktp);
      }
   else if (slr->state == SS_TWO_WAY) {
#if CHECKPKTQUEUE
      pkt_data_queue_ok(&slr->pktq, 1, 100);
#endif
      free_pkt_data(isp->pktp);
      }
   else log_msg(LOG_ERR, 0, "stream in state %d <<<<", slr->state);
   isp->pktp = NULL;
#if CHECKPKTQUEUE
   pkt_data_queue_ok(&slr->pktq, 1, 98);
#endif

   if (new) {
      stream_q_lock(&fw->sq);
      stream_push(&fw->sq, slr);
      stream_q_unlock(&fw->sq);
      }
   else {
      free_stream(isp);  /* Finished with isp */
      stream_q_lock(&fw->sq);
      tq = stream_extract(&fw->sq, slr);  /* Move slr to stream list head */
      stream_push(&fw->sq, tq);
      stream_q_unlock(&fw->sq);
      }
#if SC_DEBUG >= 3
   check_stream_chains("+++ end of each_packet", &fw->sq);
#endif
   }

void shutdown_flow_watchers(void) {
   struct flow_watcher_control *fw;
   int k;
   for (k = 0; k != n_flow_watchers; ++k) {
      fw = &flow_watcher_control_space.objects[k];
      t_semaphore_down(&fw->ispq_empty);  /* Wait for an empty entry */
      fw->eof_sp->event_type = END_OF_FILE;
      stream_q_lock(&fw->ispq);
      stream_push(&fw->ispq, fw->eof_sp);
      stream_q_unlock(&fw->ispq);
      t_semaphore_up(&fw->ispq_ready);  /* Signal a full entry */
      }
   }

void cancel_packet_watchers(void) {
   int j, r;
   for (j = 0; j != n_interfaces; ++j) {
      if (if_info[j].thread_running) {
         r = pthread_cancel(if_info[j].pw_thread);
         printf("thread %d cancel -> %d\n", j,r);  fflush(stdout);
         }
      }
   shutdown_flow_watchers();
   }

void *packet_watcher_process(void *arg) {
   struct if_info *iip = (struct if_info *)arg;
   struct timeval tv;
   uint16_t ethertype;   uint32_t hl;  libtrace_ip_t *ip;
   int key_code, fwd;
   struct stream *new_sp;
   uint32_t ts, hash;
   struct flow_watcher_control *fw;
   int fn = 0;

   fwd = nice(-5);  /* Run this thread at high priority */
      /* Result not used, keep compiler happy */

   iip->last_ts = start_ts;

   for (;;) {  /* Loop through files in list */
      while (trace_read_packet(iip->trace, iip->packet) > 0) {
         if (request_shutdown)  /* User sent SIGINT */
       	    break;  /* Test here; pthread_cancel seems to clobber iip->trace */
         iip->n_pkts += 1;  /* For av/max pkts #stats */
         ip = trace_get_layer3(iip->packet, &ethertype, &hl);
         if (!ip || (ethertype != 0x0800 && ethertype != 0x86dd))
	    n_not_ip += 1;
         else {
            tv = trace_get_timeval(iip->packet);
            ts = tv.tv_sec;
            if (ts > iip->last_ts) seconds_change(iip, ts);

            key_code = ethertype == 0x0800 ?
	       get_v4key(&new_sp, &fwd,  hl, (uint8_t *)ip)
 	       : get_v6key(&new_sp, &fwd,  hl, (uint8_t *)ip);
            if (key_code != SKEY_OK) {
               /* Couldn't get packet key (packet freed by get_key) */
               switch (key_code) {
               case SKEY_NOPEERADDR:
               case SKEY_NOTRANSADDR:
	          n_too_small += 1;
                  break;
               case SKEY_FRAGPKT:
                  n_frags += 1;
                  break;
               case SKEY_NOLOCALADDR:
                  n_no_local_address += 1;
                  break;
               case SKEY_V6OPAQUE:
                  n_v6_opaque += 1;
                  break;
               default:
                  log_msg(LOG_ERR, 66,
	             "packet_watcher: bad key_code %d !!!", key_code);
                  break;
                  }
	       }
            else {
               hash = owt_hash(new_sp);
               new_sp->fw_nbr = n_flow_watchers == 1 ? 0 :
		  // (hash >> 24) % n_flow_watchers;
                  hash % n_flow_watchers;
               fw = &flow_watcher_control_space.objects[new_sp->fw_nbr];

               t_semaphore_down(&fw->ispq_empty);  /* Wait for empty entry */
               new_sp->event_type = NO_EVENT;
               new_sp->hash = hash;
               new_sp->fwd = fwd;
               new_sp->wl = trace_get_wire_length(iip->packet);
               new_sp->ts = ts;
               new_sp->last_us = at_seconds(tv);
               stream_q_lock(&fw->ispq);
               stream_push(&fw->ispq, new_sp);
               stream_q_unlock(&fw->ispq);
               t_semaphore_up(&fw->ispq_ready);
                  /* Pass pkt_info to selected flow watcher */
   	       }
            }
         if (request_shutdown) break;  /* User sent SIGINT */
         }  /* SIGINT or trace file EOF */

      if (request_shutdown) break;  /* User sent SIGINT */
      fn += 1;
      if (fn == traces_to_process) break;  /* No more files */

      close_trace(iip);
      if (next_trace_file(iip) != 0)
         log_msg(LOG_ERR, 1, "Couldn't start next file !!!");
      log_msg(LOG_WARNING, 0, "----- opened %s", iip->if_name);
   }

   iip->thread_running = 0;
   n_interfaces_running -= 1;
   if (n_interfaces_running == 0)
      shutdown_flow_watchers();
   }

void *flow_watcher_process(void *arg) {
   struct flow_watcher_control *fw = arg;
   struct stream *isp;
   int shutdown_now, j;

   j = nice(-3);  /* Run this thread at medium priority */
      /* Result not used, keep compiler happy */
#if STARTUP_DEBUG
   printf("starting flow_watcher_process() +-+\n"); fflush(stdout); 
#endif
   for (;;) {
      t_semaphore_down(&fw->ispq_ready);  /* Wait for a full block */
      stream_q_lock(&fw->ispq);
      isp = stream_shift(&fw->ispq);
      stream_q_unlock(&fw->ispq);

      if (isp->event_type == END_OF_FILE)
         break;
      else switch (isp->event_type) {
      case S_TIME_EVENT:
#if STREAMTIMEOUTS
         mutex_lock("recording", &count_mutex);
	 if (CV_RECORDING) {  /* Handling #Record: don't check timeouts */
	    mutex_unlock("recording", &count_mutex);
	    break;
	    }
	 count += 1;  /* About to run check_stream_timeouts() */
         mutex_unlock("recording", &count_mutex);

	 check_stream_timeouts(fw, isp->ts);

	 pthread_mutex_lock(&count_mutex);
	 count -= 1;  /* check_stream_timeouts() finished */
	 pthread_cond_signal(&count_cv);  /* Let Record: know! */
	 pthread_mutex_unlock(&count_mutex);
#endif
         free_stream(isp);
	 break;
      default:
	 each_pkt(fw, isp);
	 break;
         }

      t_semaphore_up(&fw->ispq_empty);  /* Signal an empty block */
      }  /* Loop finishes on interrupt or end of trace file */

   output_active_streams(&fw->sq, isp->ts);
      /* Dump currently active streams */

   mutex_lock("fwp", &n_fw_mutex);
   n_fw_running -= 1;
   shutdown_now = n_fw_running == 0;
   mutex_unlock("fwp", &n_fw_mutex);

   if (shutdown_now) {
      for (j = 0; j != n_interfaces; ++j)
         close_trace(&if_info[j]);
      if (request_shutdown)
         stop_meter("#Shutdown: User sent interrupt ###");
      else
         stop_meter("#Shutdown: End of trace file ###");
      }
   }
