#!/usr/bin/env ruby

#  0955, Fri 26 Feb 10 (NZDT)
#  1818, Sun  9 Aug 09 (NZST)
#
#  lib/owt_monitor.rb: Class Streamer, user interface to OWTmonitor
#
#  Copyright 2007-2009, Nevil Brownlee, U Auckland | CAIDA | WAND

require 'optparse'

def streamer_usage()
  $stderr.print <<END
Usage: #{File.basename $0} [ -i interface ] [ -c interval ]
         [ -d pktdatas ] [ -s streams ]
         [ -r pktstorecord ] [ -n minpkts ] [ -t types ]
END
end

module OWTmonitor
  class Monitor
    def usage(status, msg = nil)
      $stderr.puts msg if msg
      streamer_usage
      exit(status)
    end

    attr_reader :live, :n_interfaces, :last_fn

    def check_interfaces(interfaces)
      @live = nil
      interfaces.each do |if_name|
        if if_name =~ /^(.+):(.+)$/
          scheme = $1;  @last_fn = $2
          live = !(scheme == 'pcap' ||scheme == 'pcapfile' || \
            scheme == 'erf')
          $stderr.print \
            "interface=#{if_name}, scheme=#{scheme}, live=#{live}\n"
          if @live.nil?
            @live = live
          elsif live != @live
            print "Interfaces must be all live or all !live <<<\n"
            exit
          end
        else
          $stderr.print "No : in #{if_name} (libtrace expects a URI) <<<\n"
          exit
        end
      end
      @n_interfaces = interfaces.size
    end

    def init(args = nil)
      interfaces = Array.new
      streams = 90000;  pkt_datas = 1000000
      pktstorecord = 20;  min_pkts = 10
      timerecinterval = 300;  types = 10;  hours = 1
      if args
        ARGV[0,0] = args.split(/\s+/)
           # Prepend to ARGV, so that command-line args 
           # will follow whatever we're given as a parameter
      end
      opts = OptionParser.new do |opts|
        opts.on('-c INTERVAL') { |v| timerecinterval = v.to_i }
        opts.on('-h HOURS') { |v| hours = v.to_i }
        opts.on('-i INTERFACES') { |v| interfaces.push(v) }
        opts.on('-n MINPKTS') { |v| min_pkts = v.to_i }
        opts.on('-p PKTDATAS') { |v| pkt_datas = v.to_i }
        opts.on('-r PKTSTORECORD') { |v| pktstorecord = v.to_i }
        opts.on('-s STREAMS') { |v| streams = v.to_i }
        opts.on('-t TYPES') { |v| types = v.to_i }
      end
      opts.parse!(ARGV) 
#       print "interface=#{interfaces}" \
#         + ", streams=#{streams}, pkt_datas=#{pkt_datas}" \
#            + ", timeinterval=#{timerecinterval}\n"
      if interfaces.size == 0
        $stderr.print "No interface(s) specified <<<\n"
        exit
      end
      check_interfaces(interfaces)
      self.owt_init(interfaces, @live,
        pkt_datas, streams, pktstorecord, timerecinterval,
        min_pkts, types, hours)
    end  # Monitor.init

  # Only the usage and init methods (above) are in Ruby - 
  # Monitor's other methods are in owt_monitor.c.  They are ..
  #
  #  initialize  - creates new Monitor object, calls owt_init
  #  owt_init    - does C-level intialising for init
  #  ...

  end  # class Monitor
end

Monitor = OWTmonitor::Monitor  # Allow us to use Monitor.new directly
