#!/usr/local/ruby

# 1228, Fri 4 Jun 10 (NZST)
#
# owt-support.rb: constants and utility methods for owtmon
#
# Conright 2010, Nevil Brownlee, U Auckland | CAIDA

T_unclassified    =  0  # not enough packets to classify
T_tcp_probe       =  1  # tcp only, same port, many dest addrs
T_tcp_multi_ppa   =  2  # tcp only, >1 pkt/addr
T_tcp_single_ppa  =  3  # tcp only, 1 pkt/addr
T_udp_probe       =  4  # udp only, same port, many dest addrs
T_udp_multi_ppa   =  5  # udp only, >1 pkt/addr
T_udp_single_ppa  =  6  # udp only, 1 pkt/addr
T_mixed_tcp_udp   =  7  # both tcp and udp pkts
T_no_tcp_or_udp   =  8  # only other protocols
T_backscatter     =  9  # tcp SYN+ACKs or RST+ACKs only
T_conficker_C     = 10  # only conficker p2p pkts
T_conficker_mix   = 11  # conficker p2p pkts, some app-port pkts

N_source_types    = 12  # <<< Change this when adding new types


SrcTypeNames = [ "Unclassified",  # 0
  "TCP port probe", "TCP, > 1p/a", "TCP, 1p/a", # 1, 2, 3
  "UDP port probe", "UDP, > 1p/a", "UDP, 1p/a",  # 4, 5, 6
  "TCP and UDP", "No TCP or UDP", "Backscatter",  # 7, 8, 9
  "Conficker P2P", "Conficker mixed" ]  # 10, 11

def run_command(cmd, print_output=false)  # return cmd's last output line
  $stdout.sync = true
  ll = ''
  IO.popen(cmd) do |f|
    until f.eof?
      ll = f.gets
      puts(ll) if print_output
    end
  end
  ll.chomp
end

def already_there(fn)
  return false if !File.exists?(fn)
  return false if File.size?(fn).nil?  # nil if size zero
  true
end

def check_dir_exists(dn)
  e = File.exists?(dn); d = File.directory?(dn)
  if e && !d
    print "#{date} is an existing file, not creating directory !!!\n"
    exit
  end
  Dir.mkdir(dn) if !e
end

# Telescope times are UTC, but Ruby formats them in local time!

def utc_gts(local_ts)  # Convert ts to gnuplot timefmt
  ds = local_ts.to_i+20  # strfmtime will round down to closest hour
  Time.at(ds).gmtime.strftime("%Y%m%d.%H%M")
    # Assume that gmtime uses TZ environment variable
    # time format suitable for gnuplot !
end

def utc_hts(local_ts)  # Convert ts to Ymd date with hyphens
  ds = local_ts.to_i+20  # strfmtime will round down to closest hour
  Time.at(ds).gmtime.strftime("%Y-%m-%d")
end
